SystemOrganization addCategory: #LogarithmicInterval!

!Number methodsFor: '*LogarithmicInterval' stamp: 'nice 2/8/2019 20:28'!
logTo: logStop
	"Create a logarithmic interval given by Neperian logarithm bounds.
	The ratio of geometric progression will be 1 exp"
	^LogarithmicInterval logFrom: self to: logStop by: 1 base: nil! !

!Number methodsFor: '*LogarithmicInterval' stamp: 'nice 2/8/2019 20:27'!
logTo: logStop base: aNumberOrNil
	"Create a logarithmic interval given by logarithm bounds in base aNumberOrNil.
	If base is nil, use Neperian logarithm."
	^LogarithmicInterval logFrom: self to: logStop by: 1 base: aNumberOrNil! !

!Number methodsFor: '*LogarithmicInterval' stamp: 'nice 2/8/2019 19:44'!
logTo: logStop by: logStep
	"Create a logarithmic interval given by Neperian logarithm bounds and steps"
	^LogarithmicInterval logFrom: self to: logStop by: logStep base: nil! !

!Number methodsFor: '*LogarithmicInterval' stamp: 'nice 2/8/2019 19:45'!
logTo: logStop by: logStep base: aNumberOrNil
	"Create a logarithmic interval given by logarithm bounds and steps in base aNumberOrNil.
	If base is nil, use Neperian logarithm."
	^LogarithmicInterval logFrom: self to: logStop by: logStep base: aNumberOrNil! !

!Number methodsFor: '*LogarithmicInterval' stamp: 'nice 2/8/2019 19:46'!
logTo: logStop size: anInteger
	"Create a logarithmic interval of prescribed size given by Neperian logarithm bounds"
	^LogarithmicInterval logFrom: self to: logStop size: anInteger base: nil! !

!Number methodsFor: '*LogarithmicInterval' stamp: 'nice 2/8/2019 19:46'!
logTo: logStop size: anInteger base: aNumberOrNil
	"Create a logarithmic interval of prescribed size given by logarithm bounds in base aNumberOrNil.
	If base is nil, use Neperian logarithm."
	^LogarithmicInterval logFrom: self to: logStop size: anInteger base: aNumberOrNil! !

SequenceableCollection subclass: #LogarithmicInterval
	instanceVariableNames: 'logInterval base'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'LogarithmicInterval'!

!LogarithmicInterval commentStamp: 'nice 2/8/2019 22:58' prior: 0!
A LogarithmicInterval is a kind of Interval whose elements follow a geometric progression.
A geometric serie has a constant ratio between consecutive elements.
If we note a the first element, and b this constant ratio, a geometric serie is of the kind:
    {a. a*b. a*b*b. a*(b raisedTo: 3). ...}
Or in other words, taking the logarithm of the elements of a LogarithmicInterval gives a regular arithmetic Interval of step (b log):
    {a log. b log + a log. b log * 2 + a log. b log * 3 + a log. ...}

LogarithmicInterval are created by specifying the logarithm of their bounds in whatever base:
	(0 logTo: 3 base: 2) asArray will give {2 raisedTo: 0. 2 raisedTo: 1. 2 raisedTo: 2. 2 raisedTo: 3}, that is #(1 2 4 8).

Negative and fractional power are allowed, and this will most usually lead to logarithmic intervals of Float:
	(-1/2 logTo: 1/2 base: 2) asArray will give  #(0.7071067811865476 1.414213562373095).

A logarithmic step can also be specified:
	(0 logTo: 4 by: 2 base: 2) asArray will give {2 raisedTo: 0. 2 raisedTo: 2. 2 raisedTo: 4}, that is #(1 4 16).
The geometric progression (or ratio) is then (base raisedTo: logInterval step)

The size of the LogarithmicInterval can be specified instead of the logarithmic step:
	(0 logTo: 4 size: 5 base: 3) asArray  will give #(1 3 9 27 81)

This is mostly convenient for floating point logarithmic intervals:
	(-1 logTo: 2 size: 20 base: 100) asArray.

Note that the base is not that important once we specify the size, we can also use ln and not specify the base:
	(0.1 log logTo: 10 log size: 20 base: 10) asArray.
	((0.1 log: 3) logTo: (10 log:3) size: 20 base: 3) asArray.
	(0.1 ln logTo: 10.0 ln size: 20) asArray.
Above expressions should give approximately the same LogarithmicInterval from 0.1 to 10 (modulo the floating point rounding errors).

Instance Variables
	base:		<Number | nil> the base of logarithm in which the logInterval is expressed
	logInterval:		<Interval> the logarithm of elements (in base base)

Note: if the base is nil, then logInterval is given in Neperian logarithm (see #ln) and elements are reconstructed using exponential (see #exp).
!

!LogarithmicInterval class methodsFor: 'instance creation' stamp: 'nice 2/8/2019 20:30'!
logFrom: logStart to: logStop by: logStep base: numberOrNil
	^self basicNew setLogInterval: (logStart to: logStop by: logStep) base: numberOrNil! !

!LogarithmicInterval class methodsFor: 'instance creation' stamp: 'nice 2/8/2019 19:43'!
logFrom: logStart to: logStop size: anInteger base: numberOrNil
	"Create a logarithmic interval of prescribed size"
	
	^self basicNew setLogFrom: logStart to: logStop size: anInteger base: numberOrNil ! !

!LogarithmicInterval methodsFor: 'comparing' stamp: 'nice 2/8/2019 18:45'!
= aLogaritmicInterval 
	"Answer true if aLogaritmicInterval is of my class, and has same elements.
	It should have same elements if it has same first, last and size. "
	
	^ aLogaritmicInterval class = self class
		ifTrue: [self size = aLogaritmicInterval size
				and: [self size = 0
						or: [self first = aLogaritmicInterval first
								and: [self last = aLogaritmicInterval last]]]]
		ifFalse: [super = aLogaritmicInterval]! !

!LogarithmicInterval methodsFor: 'adding' stamp: 'nice 2/8/2019 18:38'!
add: newObject 
	"Provide an error notification that adding to a LogarithmicInterval is not allowed."

	self shouldNotImplement! !

!LogarithmicInterval methodsFor: 'accessing' stamp: 'nice 2/8/2019 18:37'!
at: anInteger
	^base ifNil: [(logInterval at: anInteger) exp] ifNotNil: [base raisedTo: (logInterval at: anInteger)]! !

!LogarithmicInterval methodsFor: 'accessing' stamp: 'nice 2/8/2019 18:39'!
at: anInteger put: anObject 
	"Provide an error notification that storing into a LogarithmicInterval is not 
	allowed. "

	self error: 'you can not store into a logarithmic interval'! !

!LogarithmicInterval methodsFor: 'enumerating' stamp: 'nice 2/8/2019 19:20'!
collect: aBlock 
	^base
		ifNil: [logInterval collect: [:ln | aBlock value: ln exp]]
		ifNotNil: [logInterval collect: [:log | aBlock value: (base raisedTo: log)]]! !

!LogarithmicInterval methodsFor: 'copying' stamp: 'nice 2/8/2019 19:17'!
copyFrom: startIndex to: stopIndex
	^self class basicNew setLogInterval: (logInterval copyFrom: startIndex to: stopIndex) base: base! !

!LogarithmicInterval methodsFor: 'enumerating' stamp: 'nice 2/8/2019 19:19'!
do: aBlock 
	"Evaluate aBlock with each of the receiver's elements as the argument."

	base
		ifNil: [logInterval do: [:ln | aBlock value: ln exp]]
		ifNotNil: [logInterval do: [:log | aBlock value: (base raisedTo: log)]]! !

!LogarithmicInterval methodsFor: 'comparing' stamp: 'nice 2/8/2019 20:46'!
hash
	self size = 0 ifTrue: [^self size hashMultiply].
	^(self size hashMultiply
		bitXor: self first hash) hashMultiply
		bitXor: self last hash! !

!LogarithmicInterval methodsFor: 'testing' stamp: 'nice 2/10/2019 11:54'!
includes: aNumber
	"Provide O(1) implementation rather than super O(size)."

	^(self indexOf: aNumber startingAt: 1) > 0! !

!LogarithmicInterval methodsFor: 'accessing' stamp: 'nice 2/10/2019 11:53'!
indexOf: aNumber startingAt: startIndex
	"Answer at which index is stored aNumber.
	Answer 0 if aNumber is not in this collection or is store before startIndex.
	Provide O(1) implementation rather than super O(size)."

	| index |
	aNumber isNumber ifFalse: [^0].
	aNumber > 0 ifFalse: [^0].
	index := ((base ifNil: [aNumber ln] ifNotNil: [aNumber log: base]) - logInterval start / logInterval increment + 1) rounded.
	(index between: startIndex and: self size) ifFalse: [^0].
	(self at: index) = aNumber ifFalse: [^0].
	^index! !

!LogarithmicInterval methodsFor: 'printing' stamp: 'nice 2/8/2019 21:08'!
printOn: aStream 
	"Append to the argument aStream a sequence of characters 
	that identifies the logarithmic interval.
	(logStart logTo: logStop by: logStep base: base)"
	
	aStream nextPut: $(.
	logInterval printLogOn: aStream.
	base
		ifNotNil:
			[aStream nextPutAll: ' base: '.
			base printOn: aStream].
	aStream nextPut: $)! !

!LogarithmicInterval methodsFor: 'accessing' stamp: 'nice 2/8/2019 18:41'!
ratio
	^base ifNil: [logInterval step exp] ifNotNil: [base raisedTo: logInterval step]! !

!LogarithmicInterval methodsFor: 'converting' stamp: 'nice 2/8/2019 19:17'!
reversed
	^self class basicNew setLogInterval: logInterval reversed base: base! !

!LogarithmicInterval methodsFor: 'private' stamp: 'nice 2/8/2019 21:14'!
setLogFrom: logStart to: logStop size: anInteger base: numberOrNil 
	base := numberOrNil.
	logInterval := Interval from: logStart to: logStop size: anInteger! !

!LogarithmicInterval methodsFor: 'private' stamp: 'nice 2/8/2019 19:15'!
setLogInterval: anInterval base: aNumberOrNil
	logInterval := anInterval.
	base := aNumberOrNil! !

!LogarithmicInterval methodsFor: 'accessing' stamp: 'nice 2/8/2019 18:41'!
size
	^logInterval size! !

!LogarithmicInterval methodsFor: 'sorting' stamp: 'nice 2/10/2019 10:28'!
sorted
	"a LogarithmicInterval is already sorted"
	^logInterval increment > 0
		ifTrue: [self]
		ifFalse: [self reversed]! !

!LogarithmicInterval methodsFor: 'private' stamp: 'nice 2/8/2019 19:21'!
species
	^logInterval species! !

!LogarithmicInterval methodsFor: 'printing' stamp: 'nice 2/8/2019 21:07'!
storeOn: aStream 
	"Append to the argument aStream a sequence of characters 
	that identifies the logarithmic interval.
	(logStart logTo: logStop by: logStep base: base)"
	
	aStream nextPut: $(.
	logInterval storeLogOn: aStream.
	base
		ifNotNil:
			[aStream nextPutAll: ' base: '.
			base storeOn: aStream].
	aStream nextPut: $)! !

!Interval class methodsFor: '*LogarithmicInterval' stamp: 'nice 2/8/2019 21:14'!
from: start to: stop size: anInteger
	"Attempt to provide an interval of prescribed size despite possible float round-off errors"
	| step interval |
	anInteger = 0 ifTrue: [^ 1 to: 0 ].
	anInteger = 1 ifTrue: [^ start to: start ].
	start = stop ifTrue: [self error: 'an Interval having same start and stop bounds cannot have size > 1'].
	step := stop - start / (anInteger - 1).
	interval := start to: stop by: step.
	interval size < anInteger
		ifTrue:
			[step := step < 0
				ifTrue: [step successor]
				ifFalse: [step predecessor].
			^start to: stop by: step].
	^interval! !

!Interval methodsFor: '*LogarithmicInterval' stamp: 'nice 2/8/2019 21:07'!
printLogOn: aStream 
	"private - used for printing logarithmic intervals"
	start printOn: aStream.
	aStream nextPutAll: ' logTo: '.
	stop printOn: aStream.
	aStream nextPutAll: ' by: '.
	step printOn: aStream! !

!Interval methodsFor: '*LogarithmicInterval' stamp: 'nice 2/8/2019 21:07'!
storeLogOn: aStream 
	"private - used for storing logarithmic intervals"
	start storeOn: aStream.
	aStream nextPutAll: ' logTo: '.
	stop storeOn: aStream.
	aStream nextPutAll: ' by: '.
	step storeOn: aStream! !
